from contextlib import contextmanager

import logging
import re

import functools
from collections import namedtuple
from typing import Dict, Union, List

import ldap

from ldap import modlist
from ldap.controls import SimplePagedResultsControl
from ldap.ldapobject import SimpleLDAPObject as LdapObject

from . import exceptions

ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)

logger = logging.getLogger(__name__)


def _encode_attributes(attr_dict: dict) -> Dict[str, Union[List, bytes]]:
    @functools.singledispatch
    def _encode(value):
        raise exceptions.ParameterError(
            f"Received LDAP attribute value {value} of unsupported type "
            f"{type(value)}. Values must be one of [int, str, bytes, list]."
        )

    @_encode.register
    def _int(value: int):
        return _encode(str(value))

    @_encode.register
    def _str(value: str):
        return bytes(value, "utf-8")

    @_encode.register
    def _bytes(value: bytes):
        return value

    @_encode.register
    def _list(value: list):
        return [_encode(v) for v in value]

    @_encode.register
    def _none(_: None):
        raise exceptions.ParameterError("None values cannot be supplied when creating an object.")

    return {key: _encode(value) for key, value in attr_dict.items()}


def _process_ldap_exception(exception):
    """ Attempt to unpack python-ldap exception and return a
    namedtuple of error information. Log error and return None
    if unpacking fails.
    """
    exception_tuple = namedtuple("ldap_exception", ["error", "data"])
    # We only need to inspect the first error code and the data code. Everything
    # else can be ignored.
    regex = re.compile(r"(?P<error>\w+):.*, data (?P<data>\w+)")
    try:
        ldap_error = next(iter(exception.args))
        info = ldap_error["info"].replace("\n", "")  # sometimes they contain line breaks
    except (StopIteration, KeyError):
        # Unexpected exception format, cannot unpack.
        logger.error(f"Unable to unpack exception: {str(exception)}")
        return None
    try:
        return exception_tuple(**re.search(regex, info).groupdict())
    except AttributeError:
        # Unexpected info format, cannot parse.
        logger.error(f"Cannot parse exception info string '{info}'")
    return None


@contextmanager
def _get_session(address: str, port: Union[str, int], domain: str, username: str, password: str) -> LdapObject:
    # Without this, the `finally` block would error with:
    # UnboundLocalError: local variable 'connection' referenced before assignment
    connection = None
    try:
        connection = ldap.initialize(f"ldaps://{address}:{port}")
        connection.set_option(ldap.OPT_REFERRALS, 0)
        connection.set_option(ldap.OPT_NETWORK_TIMEOUT, 30.0)
        connection.protocol_version = ldap.VERSION3
        connection.simple_bind_s(f"{username}@{domain}", password)
        setattr(connection, "_base_dn", ",".join(list(map(lambda x: "DC={0}".format(x), domain.split(".")))))
        yield connection

    except ldap.SERVER_DOWN:
        raise exceptions.ServerDown(f"Bind failed. LDAPS server {address}:{port} is down.")

    except ldap.INVALID_CREDENTIALS as e:
        exception_data = _process_ldap_exception(e)
        try:
            raise {
                "52e": exceptions.InvalidCredentials(
                    f"Login failed for user account {username}. Invalid username or password."
                ),
                "530": exceptions.LogonNotPermitted(f"User account {username} is not permitted to logon at this time."),
                "532": exceptions.PasswordExpired(f"Password for user account {username} has expired."),
                "533": exceptions.UserAccountDisabled(f"User account {username} is disabled."),
                "568": exceptions.InvalidCredentials(
                    f"Security context for user account {username} contains too many IDs. The user account may "
                    "be in too many Security Groups."
                ),
                "701": exceptions.UserAccountExpired(f"User account {username} has expired."),
                "773": exceptions.PasswordResetRequired(f"Password for user account {username} must be reset."),
                "775": exceptions.UserAccountLocked(f"User account {username} is locked."),
            }[exception_data.data]
        except (AttributeError, KeyError):  # Either exception_data is None or not within the known keys above
            raise exceptions.InvalidCredentials(
                f"Login failed for user account {username} with an "
                f"unexpected error. Please view logs for more information. {str(e)}."
            )

    finally:
        if connection:
            connection.unbind()


def _search(connection: LdapObject, query: str, scope: int = ldap.SCOPE_SUBTREE, page_size: int = 1024) -> List[dict]:
    logger.debug(f"search query: {query}")

    def _paged_search():
        """See https://www.python-ldap.org/en/latest/reference/ldap.html
        :return: None or yield (dn, attrs) result tuples
        """
        control = SimplePagedResultsControl(size=page_size, cookie="")
        query_id = connection.search_ext(connection._base_dn, scope, query, serverctrls=[control])

        while True:
            logger.debug("getting results page")
            _code, results, _result_id, server_controls = connection.result3(query_id)

            if not results:
                return

            for result in results:
                dn, _record = result
                if dn:
                    yield result

            try:
                (page_control,) = [o for o in server_controls if isinstance(o, SimplePagedResultsControl)]
            except ValueError:
                raise BufferError("Unable to create paged results (RFC 2696)")

            if not page_control.cookie:
                return

            control.cookie = page_control.cookie

            query_id = connection.search_ext(connection._base_dn, scope, query, serverctrls=[control])

    try:
        # Strip None referrals from the response. Tuple format: (distinguishedName, attribute_dictionary).
        return [r[1] for r in _paged_search() if r[0]]
    except ldap.FILTER_ERROR:
        raise exceptions.InvalidFilter(query)
    except ldap.OPERATIONS_ERROR as e:
        exception_data = _process_ldap_exception(e)
        if exception_data and exception_data.error == "000004DC":
            raise exceptions.UserAccountDisabled(f"The user account performing the operation is disabled.")
        raise e


def _create(connection: LdapObject, distinguished_name: str, attr_dict: dict):
    mod_list = modlist.addModlist(attr_dict)
    logger.info(f"creating object {distinguished_name} with attrs {str(mod_list)}")
    try:
        return connection.add_s(distinguished_name, mod_list)
    except ldap.ALREADY_EXISTS:
        raise exceptions.ObjectExists(f"Object with distinguishedName '{distinguished_name}' already exists.")
    except ldap.NO_SUCH_ATTRIBUTE:
        raise exceptions.ParameterError(f"Tried to create an object with an invalid LDAP attribute.")


def _delete(connection: LdapObject, distinguished_name: str):
    logger.info(f"deleting object {distinguished_name}")
    return connection.delete_s(distinguished_name)


def _ldap_modify(func):
    """
    A decorator for modify_add, modify_replace, and modify_delete.
    The same exceptions can be raised by all those LDAP operations
    so it makes sense to run them all through this decorator and
    catch/re-raise in one place.
    """

    @functools.wraps(func)
    def _fn(*args):
        try:
            connection, distinguished_name, key, value = args
        except ValueError:
            raise ValueError(
                "Functions decorated with _ldap_modify must have four arguments "
                "(connection, distinguished_name, key, value)."
            )

        try:
            return func(connection, distinguished_name, key, value)
        except ldap.NO_SUCH_ATTRIBUTE as e:
            exception_data = _process_ldap_exception(e)
            try:
                raise {
                    "00000057": exceptions.ModifyAttributeError(
                        f"Performing LDAP operation {func.__name__} failed. Target object '{distinguished_name}'. "
                        f"Attribute '{key}' does not exist"
                    ),
                    "00002085": exceptions.ModifyAttributeError(
                        f"Performing LDAP operation {func.__name__} failed. Target object '{distinguished_name}'. "
                        f"Value {value} does not exist in attribute '{key}'"
                    ),
                }[exception_data.error]
            except (AttributeError, KeyError):  # Either exception_data is None or not within the known keys above
                raise exceptions.ModifyAttributeError(
                    f"Performing LDAP operation '{func.__name__}' failed. Target "
                    f"object '{distinguished_name}'. Please view logs for more information. {str(e)}"
                )

        except ldap.TYPE_OR_VALUE_EXISTS:
            raise exceptions.ModifyAttributeError(
                f"Performing LDAP operation '{func.__name__}' failed. Target "
                f"object '{distinguished_name}'. Attribute '{key}' already has a value."
            )

        except ldap.OPERATIONS_ERROR as e:
            exception_data = _process_ldap_exception(e)
            if exception_data and exception_data.error == "000004DC":
                raise exceptions.UserAccountDisabled(f"The user account performing the operation is disabled.")
            raise e

    return _fn


@_ldap_modify
def _modify_add(connection: LdapObject, distinguished_name: str, key: str, value: bytes) -> None:
    logger.debug(f"adding attribute '{key}' value '{value}' to object {distinguished_name}")
    return connection.modify_s(distinguished_name, [(ldap.MOD_ADD, key, [value])])


@_ldap_modify
def _modify_replace(connection: LdapObject, distinguished_name: str, key: str, value: bytes) -> None:
    logger.debug(f"changing attribute '{key}' value to '{value}' on object {distinguished_name}")
    return connection.modify_s(distinguished_name, [(ldap.MOD_REPLACE, key, [value])])


@_ldap_modify
def _modify_delete(connection: LdapObject, distinguished_name: str, key: str, value: bytes) -> None:
    logger.debug(f"removing attribute '{key}' value '{value}' from object {distinguished_name}")
    return connection.modify_s(distinguished_name, [(ldap.MOD_DELETE, key, [value])])
